package common.nbt;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Map;
import java.util.Map.Entry;

import common.collect.Maps;

import java.util.Set;

public class NBTTagCompound extends NBTBase {
	private Map<String, NBTBase> tags = Maps.<String, NBTBase>newHashMap();

	void write(DataOutput output) throws IOException {
		for(String key : this.tags.keySet()) {
			NBTBase tag = this.tags.get(key);
			output.writeByte(tag.getId());
			if(tag.getId() != 0) {
				output.writeUTF(key);
				tag.write(output);
			}
		}
		output.writeByte(0);
	}

	void read(DataInput input, int depth, SizeTracker tracker) throws IOException {
		tracker.read(48);
		if(depth > 512)
			throw new RuntimeException("Tried to read NBT tag with too high complexity, depth > 512");
		this.tags.clear();
		byte id;
		while((id = input.readByte()) != 0) {
			String key = input.readUTF();
			tracker.read(28 + 2 * key.length());
			NBTBase tag = NBTBase.createNewByType(id);
			tag.read(input, depth + 1, tracker);
			if(this.tags.put(key, tag) != null)
				tracker.read(36);
		}
	}

	public Set<String> getKeySet() {
		return this.tags.keySet();
	}

	public byte getId() {
		return (byte)10;
	}

	public void set(String key, NBTBase value) {
		this.tags.put(key, value);
	}

	public void setTag(String key, NBTTagCompound value) {
		this.tags.put(key, value);
	}

	public void setFloatList(String key, NBTTagFloatList value) {
		this.tags.put(key, value);
	}

	public void setDoubleList(String key, NBTTagDoubleList value) {
		this.tags.put(key, value);
	}

	public void setStringList(String key, NBTTagStringList value) {
		this.tags.put(key, value);
	}

	public void setIntArrayList(String key, NBTTagIntArrayList value) {
		this.tags.put(key, value);
	}

	public void setTagList(String key, NBTTagTagList value) {
		this.tags.put(key, value);
	}

	public void setByte(String key, byte value) {
		this.tags.put(key, new NBTTagByte(value));
	}

	public void setShort(String key, short value) {
		this.tags.put(key, new NBTTagShort(value));
	}

	public void setInt(String key, int value) {
		this.tags.put(key, new NBTTagInt(value));
	}

	public void setLong(String key, long value) {
		this.tags.put(key, new NBTTagLong(value));
	}

	public void setFloat(String key, float value) {
		this.tags.put(key, new NBTTagFloat(value));
	}

	public void setDouble(String key, double value) {
		this.tags.put(key, new NBTTagDouble(value));
	}

	public void setString(String key, String value) {
		this.tags.put(key, new NBTTagString(value));
	}

	public void setByteArray(String key, byte[] value) {
		this.tags.put(key, new NBTTagByteArray(value));
	}

	public void setIntArray(String key, int[] value) {
		this.tags.put(key, new NBTTagIntArray(value));
	}

	public void setBool(String key, boolean value) {
		this.setByte(key, (byte)(value ? 1 : 0));
	}

	public NBTBase get(String key) {
		return this.tags.get(key);
	}

	public byte getTagId(String key) {
		NBTBase tag = this.tags.get(key);
		return tag != null ? tag.getId() : 0;
	}

	private boolean hasKey(String key, int type) {
		int id = this.getTagId(key);
		return id == type || (type == 99 && id >= 1 && id <= 6);
	}

	public boolean hasBool(String key) {
		return this.hasByte(key);
	}

	public boolean hasByte(String key) {
		return this.hasKey(key, 1);
	}

	public boolean hasShort(String key) {
		return this.hasKey(key, 2);
	}

	public boolean hasInt(String key) {
		return this.hasKey(key, 3);
	}

	public boolean hasLong(String key) {
		return this.hasKey(key, 4);
	}

	public boolean hasFloat(String key) {
		return this.hasKey(key, 5);
	}

	public boolean hasDouble(String key) {
		return this.hasKey(key, 6);
	}

	public boolean hasString(String key) {
		return this.hasKey(key, 8);
	}

	public boolean hasByteArray(String key) {
		return this.hasKey(key, 7);
	}

	public boolean hasIntArray(String key) {
		return this.hasKey(key, 11);
	}

	public boolean hasTag(String key) {
		return this.hasKey(key, 10);
	}

	public boolean hasFloatList(String key) {
		return this.hasKey(key, 13);
	}

	public boolean hasDoubleList(String key) {
		return this.hasKey(key, 14);
	}

	public boolean hasStringList(String key) {
		return this.hasKey(key, 9);
	}

	public boolean hasIntArrayList(String key) {
		return this.hasKey(key, 15);
	}

	public boolean hasTagList(String key) {
		return this.hasKey(key, 12);
	}

	public byte getByte(String key) {
		return !this.hasKey(key, 1) ? 0 : ((NBTTagByte)this.tags.get(key)).getByte();
	}

	public short getShort(String key) {
		return !this.hasKey(key, 2) ? 0 : ((NBTTagShort)this.tags.get(key)).getShort();
	}

	public int getInt(String key) {
		return !this.hasKey(key, 3) ? 0 : ((NBTTagInt)this.tags.get(key)).getInt();
	}

	public long getLong(String key) {
		return !this.hasKey(key, 4) ? 0L : ((NBTTagLong)this.tags.get(key)).getLong();
	}

	public float getFloat(String key) {
		return !this.hasKey(key, 5) ? 0.0F : ((NBTTagFloat)this.tags.get(key)).getFloat();
	}

	public double getDouble(String key) {
		return !this.hasKey(key, 6) ? 0.0D : ((NBTTagDouble)this.tags.get(key)).getDouble();
	}

	public String getString(String key) {
		return !this.hasKey(key, 8) ? "" : ((NBTTagString)this.tags.get(key)).getString();
	}

	public byte[] getByteArray(String key) {
		return !this.hasKey(key, 7) ? new byte[0] : ((NBTTagByteArray)this.tags.get(key)).getByteArray();
	}

	public int[] getIntArray(String key) {
		return !this.hasKey(key, 11) ? new int[0] : ((NBTTagIntArray)this.tags.get(key)).getIntArray();
	}

	public NBTTagCompound getTag(String key) {
		return !this.hasKey(key, 10) ? new NBTTagCompound() : (NBTTagCompound)this.tags.get(key);
	}

	public NBTTagFloatList getFloatList(String key) {
		return !this.hasKey(key, 13) ? new NBTTagFloatList() : (NBTTagFloatList)this.tags.get(key);
	}

	public NBTTagDoubleList getDoubleList(String key) {
		return !this.hasKey(key, 14) ? new NBTTagDoubleList() : (NBTTagDoubleList)this.tags.get(key);
	}

	public NBTTagStringList getStringList(String key) {
		return !this.hasKey(key, 9) ? new NBTTagStringList() : (NBTTagStringList)this.tags.get(key);
	}

	public NBTTagIntArrayList getIntArrayList(String key) {
		return !this.hasKey(key, 15) ? new NBTTagIntArrayList() : (NBTTagIntArrayList)this.tags.get(key);
	}

	public NBTTagTagList getTagList(String key) {
		return !this.hasKey(key, 12) ? new NBTTagTagList() : (NBTTagTagList)this.tags.get(key);
	}

	public boolean getBool(String key) {
		return this.getByte(key) != 0;
	}

	public void remove(String key) {
		this.tags.remove(key);
	}

	public String toString() {
		StringBuilder sb = new StringBuilder("{");
		for(Entry<String, NBTBase> entry : this.tags.entrySet()) {
			if(sb.length() != 1)
				sb.append(',');
			sb.append(entry.getKey()).append(':').append(entry.getValue());
		}
		return sb.append('}').toString();
	}

	public boolean isEmpty() {
		return this.tags.isEmpty();
	}

	public NBTBase copy() {
		NBTTagCompound tag = new NBTTagCompound();
		for(String s : this.tags.keySet()) {
			tag.set(s, this.tags.get(s).copy());
		}
		return tag;
	}

	public boolean equals(Object other) {
		return super.equals(other) && this.tags.entrySet().equals(((NBTTagCompound)other).tags.entrySet());
	}

	public int hashCode() {
		return super.hashCode() ^ this.tags.hashCode();
	}

	/**
	 * Merges this NBTTagCompound with the given compound. Any sub-compounds are
	 * merged using the same methods, other types of tags are overwritten from the
	 * given compound.
	 */
	public void merge(NBTTagCompound other) {
		for(String key : other.tags.keySet()) {
			NBTBase tag = other.tags.get(key);
			if(tag.getId() == 10) {
				if(this.hasKey(key, 10)) {
					NBTTagCompound comp = this.getTag(key);
					comp.merge((NBTTagCompound)tag);
				}
				else {
					this.set(key, tag.copy());
				}
			}
			else {
				this.set(key, tag.copy());
			}
		}
	}
}
